


module L = Llvm
module A = Racontrast
(* open Racontrsast *)
module StringMap = Map.Make(String)

exception NotImplementExp
exception NotImplementStmt
exception DefFuncError

let rec get_clzs prog =
  match prog with
  | [] -> []
  | (A.Class _) :: rest -> (List.hd prog) :: (get_clzs rest)
  | _ :: rest -> get_clzs rest


let rec get_noclz prog =
  match prog with
  | [] -> []
  | A.Class _ :: rest -> get_noclz rest
  | A.Function _ :: rest -> get_noclz rest
  | _ :: rest -> (List.hd prog) :: (get_noclz rest)

let rec get_funcs prog =
  match prog with
  | [] -> []
  | A.Function _ :: rest -> (List.hd prog) :: (get_funcs rest)
  | _ :: rest -> get_funcs rest




exception TypNotImpl

let translate prog =
  let context = L.global_context() in
  let the_module = L.create_module context "Racontr" in

  (* Get types from the context *)
  let i32_t      = L.i32_type    context
  and i8_t       = L.i8_type     context
  and i1_t       = L.i1_type     context
  and float_t    = L.double_type context
  and string_t   = L.pointer_type (L.i8_type context)
  and void_t     = L.void_type   context in

  let ltype_of_typ = function
      "i32"   -> i32_t
    | "string"  -> string_t
    | "bool" -> i1_t
    | "unit" -> void_t
    | _ -> raise TypNotImpl
  in



  let var_map = ref StringMap.empty in
  let add_map n v = var_map := StringMap.add n v (!var_map) in

  let lookup n = 
    StringMap.find n !var_map in

  

  let printf_t : L.lltype = 
    L.var_arg_function_type i32_t [| L.pointer_type i8_t |]
  in

  let printf_func : L.llvalue = 
    L.declare_function "printf" printf_t the_module
  in

  let clzs = get_clzs prog in
  let nonclz = get_noclz prog in




  let add_terminal builder instr =
    match L.block_terminator (L.insertion_block builder) with
      Some _ -> ()
    | None -> ignore (instr builder) in


  let all_funcs = get_funcs prog in

  let cfunc main_func body retType =
    let main_builder = L.builder_at_end context (L.entry_block main_func) in
    let int_format_str = L.build_global_stringptr "%d\n" "fmt" main_builder
    and float_format_str = L.build_global_stringptr "%g\n" "fmt" main_builder in
    let rec cexp builder e =
      match e with
      | A.Literal v -> L.const_int i32_t v
      | A.Fliteral s -> L.build_global_stringptr s s builder
      | A.BoolLit b -> L.const_int i1_t (if b then 1 else 0)
      | A.Id id -> L.build_load (lookup id) id builder
      | A.Binop (e1, op, e2) ->
          let e1' = cexp builder e1 in
          let e2' = cexp builder e2 in
            (match op with
              "+"     -> L.build_add
            | "-"     -> L.build_sub
            | "*"    -> L.build_mul
            | "/"     -> L.build_sdiv
            | "&&"     -> L.build_and
            | "||"      -> L.build_or
            | "=="   -> L.build_icmp L.Icmp.Eq
            | "!="     -> L.build_icmp L.Icmp.Ne
            | "<"    -> L.build_icmp L.Icmp.Slt
            | "<="     -> L.build_icmp L.Icmp.Sle
            | ">" -> L.build_icmp L.Icmp.Sgt
            | ">="     -> L.build_icmp L.Icmp.Sge
            ) e1' e2' "tmp" builder
      | A.Call ("print_int", [e]) ->
          L.build_call printf_func [| int_format_str ; (cexp builder e)|] "print_int" builder
      | A.Call ("print_string", [e]) ->
          L.build_call printf_func [| cexp builder e |] "print_string" builder
      | A.Call (id, args) ->
          let callee = lookup id in
            L.build_call
            callee 
            (Array.of_list (List.map (fun arg -> cexp builder arg) args))
            (id ^ "_result")
            builder
      | A.Return e -> 
        L.build_ret (cexp builder e) builder
      | _ -> raise NotImplementExp
    in
    let rec cstmt builder stmt =
      match stmt with
      | A.Var (id, e, typ) -> 
        let v = (L.build_alloca (ltype_of_typ typ) id builder) in
        let _ = add_map id v in
        let e' = cexp builder e in
        ignore(
          L.build_store e' 
          v builder
        ) ; builder
      | A.ExpStmt e -> ignore(cexp builder e) ; builder
      | A.Assign (id, e) -> 
        let e' = cexp builder e in
        ignore(
          L.build_store e' 
          (lookup id) builder
        ) ; builder

      | A.If (predicate, t, f) ->
          let bool_val = cexp builder predicate in
          let merge_bb = L.append_block context "merge" main_func in
          let build_br_merge = L.build_br merge_bb in (* partial function *)

          let then_bb = L.append_block context "then" main_func in
            add_terminal (cstmt (L.builder_at_end context then_bb) t)
              build_br_merge;

          let else_bb = L.append_block context "else" main_func in
            add_terminal (cstmt (L.builder_at_end context else_bb) f)
              build_br_merge;

          ignore(L.build_cond_br bool_val then_bb else_bb builder);
          L.builder_at_end context merge_bb
      | A.While (e, stmt) ->
          let pred_bb = L.append_block context "while" main_func in
          ignore(L.build_br pred_bb builder);

          let body_bb = L.append_block context "while_body" main_func in
          add_terminal (cstmt (L.builder_at_end context body_bb) stmt)
            (L.build_br pred_bb);

          let pred_builder = L.builder_at_end context pred_bb in
          let bool_val = cexp pred_builder e in

          let merge_bb = L.append_block context "merge" main_func in
          ignore(L.build_cond_br bool_val body_bb merge_bb pred_builder);
          L.builder_at_end context merge_bb
        
      | A.Block ss ->
          List.fold_left (fun builder s -> cstmt builder s) builder ss
        
      | _ -> raise NotImplementStmt
    in
      match body with
      | A.Block body -> 
        let builder = List.fold_left
          (fun builder stmt -> cstmt builder stmt)
          main_builder
          body
        in
        builder

      | _ -> raise DefFuncError
  in

  let def_func f = 
    match f with
    | A.Function (id, args, retTyp, body) ->
      let argTypes = Array.of_list (List.map (fun arg -> ltype_of_typ (snd arg)) args) in
      let retType = (ltype_of_typ retTyp) in
      let t = L.function_type retType argTypes  in
      let llfunc = L.define_function id t the_module in
      let builder = L.builder_at_end context (L.entry_block llfunc) in
      let () = List.iter
        (fun arg ->
          let local = L.build_alloca (ltype_of_typ (snd arg)) (fst arg) builder in
          add_map (fst arg) local
         ) args in
      let _ = List.iter
        (fun
         ((arg, argtyp), p) ->
            L.set_value_name arg p;
            ignore(L.build_store p (lookup arg) builder)
        )
        (List.combine args (Array.to_list (L.params llfunc)))
      in
      let _ = add_map id llfunc in
      cfunc llfunc body retType
    | _ -> raise DefFuncError
  in


  let _ = List.iter (fun f -> ignore(def_func f)) all_funcs in
  let main_builder = def_func (A.Function ("main", [], "i32", A.Block nonclz)) in

  let _ = add_terminal main_builder (L.build_ret (L.const_int i32_t 0)) in
    the_module
